""" View for cleaner buildout calls """
from pyramid.view import view_config
from pyramid_duh import addslash

from pypicloud.route import PackagesResource
from pypicloud.views.simple import packages_to_dict


@view_config(context=PackagesResource, request_method='GET', subpath=(),
             renderer='packages.jinja2')
@addslash
def list_packages(request):
    """ Render the list for all versions of all packages """
    names = request.db.distinct()
    # remove the ones that you are not allowed to see
    names = filter(lambda x: request.access.has_permission(x, 'read'),
                   names)
    packages = []
    for package_name in names:
        packages += request.db.all(package_name)
    return {'pkgs': packages_to_dict(request, packages)}



""" Classes that provide user and package permissions """
from functools import partial
from pyramid.path import DottedNameResolver

from .config import ConfigAccessBackend
from .base import IAccessBackend, IMutableAccessBackend, pwd_context
from .remote import RemoteAccessBackend
from .sql import SQLAccessBackend
from pypicloud.util import getdefaults


def includeme(config):
    """ Configure the app """
    settings = config.get_settings()

    resolver = DottedNameResolver(__name__)
    dotted_name = getdefaults(settings, 'pypi.auth', 'pypi.access_backend',
                              'config')
    if dotted_name == 'config':
        dotted_name = ConfigAccessBackend
    elif dotted_name == 'remote':
        dotted_name = RemoteAccessBackend
    elif dotted_name == 'sql':
        dotted_name = SQLAccessBackend
    elif dotted_name == 'ldap':
        dotted_name = "pypicloud.access.ldap_.LDAPAccessBackend"
    access_backend = resolver.maybe_resolve(dotted_name)
    kwargs = access_backend.configure(settings)
    config.add_request_method(partial(access_backend, **kwargs), name='access',
                              reify=True)



""" Commandline scripts """
import sys

import argparse
import getpass
import gzip
import json
import logging
import transaction
from base64 import b64encode
from jinja2 import Template
from pkg_resources import resource_string  # pylint: disable=E0611
from pyramid.paster import bootstrap

import os
from pypicloud.access import pwd_context


def gen_password():
    """ Generate a salted password """
    print _gen_password()


def _gen_password():
    """ Prompt user for a password twice for safety """
    while True:
        password = getpass.getpass()
        verify = getpass.getpass()
        if password == verify:
            return pwd_context.encrypt(password)
        else:
            print "Passwords do not match!"

NO_DEFAULT = object()


def prompt(msg, default=NO_DEFAULT, validate=None):
    """ Prompt user for input """
    while True:
        response = raw_input(msg + ' ').strip()
        if not response:
            if default is NO_DEFAULT:
                continue
            return default
        if validate is None or validate(response):
            return response


def prompt_option(text, choices, default=NO_DEFAULT):
    """ Prompt the user to choose one of a list of options """
    while True:
        for i, msg in enumerate(choices):
            print "[%d] %s" % (i + 1, msg)
        response = prompt(text, default=default)
        try:
            idx = int(response) - 1
            return choices[idx]
        except (ValueError, IndexError):
            print "Invalid choice\n"


def promptyn(msg, default=None):
    """ Display a blocking prompt until the user confirms """
    while True:
        yes = "Y" if default else "y"
        if default or default is None:
            no = "n"
        else:
            no = "N"
        confirm = prompt("%s [%s/%s]" % (msg, yes, no), '').lower()
        if confirm in ('y', 'yes'):
            return True
        elif confirm in ('n', 'no'):
            return False
        elif len(confirm) == 0 and default is not None:
            return default


def make_config(argv=None):
    """ Create a server config file """
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description=make_config.__doc__)
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-d', action='store_true',
                       help="Create config file for development")
    group.add_argument('-t', action='store_true',
                       help="Create config file for testing")
    group.add_argument('-p', action='store_true',
                       help="Create config file for production")
    group.add_argument('-r', action='store_true',
                       help="Create config file for docker image")

    parser.add_argument('outfile', nargs='?',
                        help="Name of output file (default stdout)")

    args = parser.parse_args(argv)

    if args.outfile is not None and os.path.exists(args.outfile):
        msg = "'%s' already exists. Overwrite?" % args.outfile
        if not promptyn(msg, False):
            return

    if args.d:
        env = 'dev'
    elif args.t:
        env = 'test'
    elif args.p:
        env = 'prod'
    elif args.r:
        env = 'docker'
    else:
        env = prompt_option("What is this config file for?",
                            ['dev', 'test', 'prod'])

    data = {
        'env': env,
    }
    data['reload_templates'] = env == 'dev'

    storage = prompt_option("Where do you want to store your packages?",
                            ['s3', 'filesystem'])
    if storage == 'filesystem':
        storage = 'file'

    data['storage'] = storage

    if storage == 's3':
        if 'AWS_ACCESS_KEY_ID' in os.environ:
            data['access_key'] = os.environ['AWS_ACCESS_KEY_ID']
        else:
            data['access_key'] = prompt("AWS access key id?")
        if 'AWS_SECRET_ACCESS_KEY' in os.environ:
            data['secret_key'] = os.environ['AWS_SECRET_ACCESS_KEY']
        else:
            data['secret_key'] = prompt("AWS secret access key?")

        def bucket_validate(name):
            """ Check for valid bucket name """
            if '.' in name:
                print "Bucket names cannot contain '.'"
                return False
            return True

        data['s3_bucket'] = prompt("S3 bucket name?", validate=bucket_validate)

    data['encrypt_key'] = b64encode(os.urandom(32))
    data['validate_key'] = b64encode(os.urandom(32))

    data['admin'] = prompt("Admin username?")
    data['password'] = _gen_password()

    data['session_secure'] = env == 'prod'
    data['env'] = env

    if env == 'dev' or env == 'test':
        data['wsgi'] = 'waitress'
    else:
        if hasattr(sys, 'real_prefix'):
            data['venv'] = sys.prefix
        data['wsgi'] = 'uwsgi'

    tmpl_str = resource_string('pypicloud', 'templates/config.ini.jinja2')
    template = Template(tmpl_str)

    config_file = template.render(**data)
    if args.outfile is None:
        sys.stdout.write(config_file)
        sys.stdout.write(os.linesep)
    else:
        with open(args.outfile, 'w') as ofile:
            ofile.write(config_file)

        print "Config file written to '%s'" % args.outfile


def migrate_packages(argv=None):
    """
    Migrate packages from one storage backend to another

    Create two config.ini files that are configured to use different storage
    backends. All packages will be migrated from the storage backend in the
    first to the storage backend in the second.

    ex: pypicloud-migrate-packages file_config.ini s3_config.ini

    """
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(
        description=migrate_packages.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('config_from',
                        help="Name of config file to migrate from")
    parser.add_argument('config_to',
                        help="Name of config file to migrate to")

    args = parser.parse_args(argv)
    logging.basicConfig()

    old_env = bootstrap(args.config_from)

    old_storage = old_env['request'].db.storage

    all_packages = old_storage.list()

    new_env = bootstrap(args.config_to)
    new_storage = new_env['request'].db.storage
    for package in all_packages:
        print "Migrating %s" % package
        with old_storage.open(package) as data:
            # we need to recalculate the path for the new storage config
            package.data.pop('path', None)
            new_storage.upload(package, data)


def export_access(argv=None):
    """ Dump the access control data to a universal format """
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description=export_access.__doc__)
    parser.add_argument('config', help="Name of config file")
    parser.add_argument('-o', help="Name of output file")

    args = parser.parse_args(argv)
    logging.basicConfig()

    env = bootstrap(args.config)
    access = env['request'].access
    data = access.dump()
    if args.o:
        with gzip.open(args.o, 'w') as ofile:
            json.dump(data, ofile)
    else:
        print json.dumps(data, indent=2)


def import_access(argv=None):
    """
    Load the access control data from a dump file or stdin

    This operation is idempotent and graceful. It will not clobber your
    existing ACL.

    """
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(
        description=import_access.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('config', help="Name of config file")
    parser.add_argument('-i', help="Name of input file")

    args = parser.parse_args(argv)
    logging.basicConfig()

    if args.i:
        with gzip.open(args.i, 'r') as ifile:
            data = json.load(ifile)
    else:
        print "Reading data from stdin..."
        data = json.load(sys.stdin)

    env = bootstrap(args.config)
    access = env['request'].access
    result = access.load(data)
    transaction.commit()
    if result is not None:
        print result



""" Access backend for storing permissions in using SQLAlchemy """
from sqlalchemy import (engine_from_config, Column, Text, Boolean, Table,
                        ForeignKey)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, backref
from sqlalchemy import orm
# pylint: disable=F0401,E0611
from zope.sqlalchemy import ZopeTransactionExtension
# pylint: enable=F0401,E0611

from .base import IMutableAccessBackend


# pylint: disable=C0103,W0231
Base = declarative_base()

association_table = Table(
    'pypicloud_user_groups', Base.metadata,
    Column('username', Text(), ForeignKey('pypicloud_users.username'), primary_key=True),
    Column('group', Text(), ForeignKey('pypicloud_groups.name'), primary_key=True)
)
# pylint: enable=C0103


class KeyVal(Base):

    """ Simple model for storing key-value pairs """
    __tablename__ = 'pypicloud_keyvals'
    key = Column(Text(), primary_key=True)
    value = Column(Text())

    def __init__(self, key, value):
        self.key = key
        self.value = value


class User(Base):

    """ User record """
    __tablename__ = 'pypicloud_users'
    username = Column(Text(), primary_key=True)
    password = Column('password', Text(), nullable=False)
    admin = Column(Boolean(), nullable=False)
    pending = Column(Boolean(), nullable=False)
    groups = orm.relationship('Group', secondary=association_table,
                              cascade='all', collection_class=set,
                              backref=backref('users', collection_class=set))

    def __init__(self, username, password, pending=True):
        self.username = username
        self.password = password
        self.groups = set()
        self.permissions = []
        self.admin = False
        self.pending = pending


class Group(Base):

    """ Group record """
    __tablename__ = 'pypicloud_groups'
    name = Column(Text(), primary_key=True)

    def __init__(self, name):
        self.name = name
        self.users = set()
        self.permissions = []


class Permission(Base):

    """ Base class for user and group permissions """
    __abstract__ = True
    package = Column(Text(), primary_key=True)
    read = Column(Boolean())
    write = Column(Boolean())

    def __init__(self, package, read, write):
        self.package = package
        self.read = read
        self.write = write

    @property
    def permissions(self):
        """ Construct permissions list """
        perms = []
        if self.read:
            perms.append('read')
        if self.write:
            perms.append('write')
        return perms

# pylint: disable=E1002


class UserPermission(Permission):

    """ Permissions for a user on a package """
    __tablename__ = 'pypicloud_user_permissions'
    username = Column(Text(), ForeignKey(User.username), primary_key=True)
    user = orm.relationship("User",
                            backref=backref('permissions',
                                            cascade='all, delete-orphan'))

    def __init__(self, package, username, read=False, write=False):
        super(UserPermission, self).__init__(package, read, write)
        self.username = username


class GroupPermission(Permission):

    """ Permissions for a group on a package """
    __tablename__ = 'pypicloud_group_permissions'
    groupname = Column(Text(), ForeignKey(Group.name), primary_key=True)
    group = orm.relationship("Group",
                             backref=backref('permissions',
                                             cascade='all, delete-orphan'))

    def __init__(self, package, groupname, read=False, write=False):
        super(GroupPermission, self).__init__(package, read, write)
        self.groupname = groupname


class SQLAccessBackend(IMutableAccessBackend):

    """
    This backend allows you to store all user and package permissions in a SQL
    database

    """

    def __init__(self, request=None, dbmaker=None, **kwargs):
        super(SQLAccessBackend, self).__init__(request, **kwargs)
        self.db = dbmaker()

        def cleanup(_):
            """ Close the session after the request """
            self.db.close()
        request.add_finished_callback(cleanup)

    @classmethod
    def configure(cls, settings):
        kwargs = super(SQLAccessBackend, cls).configure(settings)
        engine = engine_from_config(settings, prefix='auth.db.')
        kwargs['dbmaker'] = sessionmaker(
            bind=engine, extension=ZopeTransactionExtension())
        # Create SQL schema if not exists
        Base.metadata.create_all(bind=engine)
        return kwargs

    def allow_register(self):
        ret = self.db.query(KeyVal).filter_by(key='allow_register').first()
        return ret is None

    def set_allow_register(self, allow):
        if allow:
            self.db.query(KeyVal).filter_by(key='allow_register').delete()
        else:
            k = KeyVal('allow_register', 'false')
            self.db.add(k)

    def _get_password_hash(self, username):
        user = self.db.query(User).filter_by(username=username).first()
        if user:
            return user.password

    def groups(self, username=None):
        if username is None:
            query = self.db.query(Group)
            return [g.name for g in query]
        else:
            user = self.db.query(User).filter_by(username=username).first()
            if user is None:
                return []
            return [g.name for g in user.groups]

    def group_members(self, group):
        g = self.db.query(Group).filter_by(name=group).first()
        if not g:
            return []
        return [u.username for u in g.users]

    def is_admin(self, username):
        user = self.db.query(User).filter_by(username=username).first()
        return user and user.admin

    def group_permissions(self, package, group=None):
        if group is None:
            query = self.db.query(GroupPermission).filter_by(package=package)
            perms = {}
            for perm in query:
                perms[perm.groupname] = perm.permissions
            return perms
        else:
            perm = self.db.query(GroupPermission)\
                .filter_by(package=package, groupname=group).first()
            if perm:
                return perm.permissions
            return []

    def user_permissions(self, package, username=None):
        if username is None:
            query = self.db.query(UserPermission).filter_by(package=package)
            perms = {}
            for perm in query:
                perms[perm.username] = perm.permissions
            return perms
        else:
            perm = self.db.query(UserPermission)\
                .filter_by(package=package, username=username).first()
            if perm:
                return perm.permissions
            return []

    def user_package_permissions(self, username):
        query = self.db.query(UserPermission).filter_by(username=username)
        packages = []
        for perm in query:
            packages.append({
                'package': perm.package,
                'permissions': perm.permissions,
            })
        return packages

    def group_package_permissions(self, group):
        query = self.db.query(GroupPermission).filter_by(groupname=group)
        packages = []
        for perm in query:
            packages.append({
                'package': perm.package,
                'permissions': perm.permissions,
            })
        return packages

    def user_data(self, username=None):
        if username is None:
            query = self.db.query(User).filter_by(pending=False)
            users = []
            for user in query:
                users.append({
                    'username': user.username,
                    'admin': user.admin,
                })
            return users
        else:
            user = self.db.query(User).filter_by(username=username,
                                                 pending=False).first()
            if user is not None:
                return {
                    'username': user.username,
                    'admin': user.admin,
                    'groups': [g.name for g in user.groups],
                }

    def need_admin(self):
        return self.db.query(User).filter_by(admin=True).first() is None

    def _register(self, username, password):
        user = User(username, password)
        self.db.add(user)

    def pending_users(self):
        query = self.db.query(User).filter_by(pending=True)
        return [u.username for u in query]

    def approve_user(self, username):
        user = self.db.query(User).filter_by(username=username).first()
        if user is not None:
            user.pending = False

    def _set_password_hash(self, username, password_hash):
        user = self.db.query(User).filter_by(username=username).first()
        if user is not None:
            user.password = password_hash

    def delete_user(self, username):
        self.db.query(User).filter_by(username=username).delete()
        clause = association_table.c.username == username
        self.db.execute(association_table.delete(clause))

    def set_user_admin(self, username, admin):
        user = self.db.query(User).filter_by(username=username).first()
        if user is not None:
            user.admin = admin

    def edit_user_group(self, username, groupname, add):
        user = self.db.query(User).filter_by(username=username).first()
        group = self.db.query(Group).filter_by(name=groupname).first()
        if user is not None and group is not None:
            if add:
                user.groups.add(group)
            else:
                user.groups.remove(group)

    def create_group(self, group):
        self.db.add(Group(group))

    def delete_group(self, group):
        self.db.query(Group).filter_by(name=group).delete()
        clause = association_table.c.group == group
        self.db.execute(association_table.delete(clause))

    def edit_user_permission(self, package, username, perm, add):
        record = self.db.query(UserPermission)\
            .filter_by(package=package, username=username).first()
        if record is None:
            if not add:
                return
            record = UserPermission(package, username)
            self.db.add(record)
        if perm == 'read':
            record.read = add
        elif perm == 'write':
            record.write = add
        else:
            raise ValueError("Unrecognized permission '%s'" % perm)
        if not record.read and not record.write:
            self.db.delete(record)

    def edit_group_permission(self, package, group, perm, add):
        record = self.db.query(GroupPermission)\
            .filter_by(package=package, groupname=group).first()
        if record is None:
            if not add:
                return
            record = GroupPermission(package, group)
            self.db.add(record)
        if perm == 'read':
            record.read = add
        elif perm == 'write':
            record.write = add
        else:
            raise ValueError("Unrecognized permission '%s'" % perm)
        if not record.read and not record.write:
            self.db.delete(record)



""" Backend that defers to another server for access control """
from .base import IAccessBackend


class RemoteAccessBackend(IAccessBackend):

    """
    This backend allows you to defer all user auth and permissions to a remote
    server. It requires the ``requests`` package.

    """

    def __init__(self, request=None, settings=None, server=None, auth=None,
                 **kwargs):
        super(RemoteAccessBackend, self).__init__(request, **kwargs)
        self._settings = settings
        self.server = server
        self.auth = auth

    @classmethod
    def configure(cls, settings):
        kwargs = super(RemoteAccessBackend, cls).configure(settings)
        kwargs['settings'] = settings
        kwargs['server'] = settings['auth.backend_server']
        auth = None
        user = settings.get('auth.user')
        if user is not None:
            password = settings.get('auth.password')
            auth = (user, password)
        kwargs['auth'] = auth
        return kwargs

    def _req(self, uri, params=None):
        """ Hit a server endpoint and return the json response """
        try:
            import requests
        except ImportError:  # pragma: no cover
            raise ImportError("You must 'pip install requests' before using "
                              "the remote server access backend")
        response = requests.get(self.server + uri, params=params,
                                auth=self.auth)
        response.raise_for_status()
        return response.json()

    def verify_user(self, username, password):
        uri = self._settings.get('auth.uri.verify', '/verify')
        params = {'username': username, 'password': password}
        return self._req(uri, params)

    def _get_password_hash(self, username):
        # We don't have to do anything here because we overrode 'verify_user'
        pass

    def groups(self, username=None):
        uri = self._settings.get('auth.uri.groups', '/groups')
        params = {}
        if username is not None:
            params['username'] = username
        return self._req(uri, params)

    def group_members(self, group):
        uri = self._settings.get('auth.uri.group_members', '/group_members')
        params = {'group': group}
        return self._req(uri, params)

    def is_admin(self, username):
        uri = self._settings.get('auth.uri.admin', '/admin')
        params = {'username': username}
        return self._req(uri, params)

    def group_permissions(self, package, group=None):
        uri = self._settings.get('auth.uri.group_permissions',
                                 '/group_permissions')
        params = {'package': package}
        if group is not None:
            params['group'] = group
        return self._req(uri, params)

    def user_permissions(self, package, username=None):
        uri = self._settings.get('auth.uri.user_permissions',
                                 '/user_permissions')
        params = {'package': package}
        if username is not None:
            params['username'] = username
        return self._req(uri, params)

    def user_package_permissions(self, username):
        uri = self._settings.get('auth.uri.user_package_permissions',
                                 '/user_package_permissions')
        params = {'username': username}
        return self._req(uri, params)

    def group_package_permissions(self, group):
        uri = self._settings.get('auth.uri.group_package_permissions',
                                 '/group_package_permissions')
        params = {'group': group}
        return self._req(uri, params)

    def user_data(self, username=None):
        uri = self._settings.get('auth.uri.user_data',
                                 '/user_data')
        params = None
        if username is not None:
            params = {'username': username}
        return self._req(uri, params)



""" S3-backed pypi server """
import os
import sys
import traceback
import datetime

import logging
from pyramid.config import Configurator
from pyramid.renderers import JSON, render
from pyramid.settings import asbool
from pyramid_beaker import session_factory_from_settings
from six.moves.urllib.parse import urlencode  # pylint: disable=F0401,E0611

from .route import Root
from .util import BetterScrapingLocator


__version__ = '0.4.3'
LOG = logging.getLogger(__name__)


def to_json(value):
    """ A json filter for jinja2 """
    return render('json', value)

json_renderer = JSON()  # pylint: disable=C0103
json_renderer.add_adapter(datetime.datetime, lambda obj, r:
                          float(obj.strftime('%s.%f')))


def _app_url(request, *paths, **params):
    """ Get the base url for the root of the app plus an optional path """
    path = '/'.join(paths)
    if not path.startswith('/'):
        path = '/' + path
    if params:
        path += '?' + urlencode(params)
    return request.application_url + path


def _locator(request):
    """ Get the scraping locator to find packages from the fallback site """
    return BetterScrapingLocator(request.registry.fallback_url)


def includeme(config):
    """ Set up and configure the pypicloud app """
    config.set_root_factory(Root)
    config.add_route('health', '/health')
    config.include('pyramid_tm')
    config.include('pyramid_beaker')
    config.include('pyramid_duh')
    config.include('pyramid_duh.auth')
    config.include('pypicloud.auth')
    config.include('pypicloud.access')
    config.include('pypicloud.cache')
    settings = config.get_settings()

    config.add_renderer('json', json_renderer)
    # Jinja2 configuration
    settings['jinja2.filters'] = {
        'static_url': 'pyramid_jinja2.filters:static_url_filter',
        'tojson': to_json,
    }
    settings['jinja2.directories'] = ['pypicloud:templates']
    config.include('pyramid_jinja2')

    # BEAKER CONFIGURATION
    config.registry.secure_cookie = asbool(settings.get('session.secure',
                                                        False))
    settings.setdefault('session.type', 'cookie')
    settings.setdefault('session.httponly', 'true')
    config.set_session_factory(session_factory_from_settings(settings))

    # PYPICLOUD SETTINGS
    default_url = 'https://pypi.python.org/simple'
    config.registry.fallback_url = settings.get('pypi.fallback_url',
                                                default_url)

    fallback_mode = settings.get('pypi.fallback', 'redirect')
    # Compatibility with the deprecated pypi.use_fallback option
    if 'pypi.fallback' not in settings and 'pypi.use_fallback' in settings:
        LOG.warn("Using deprecated option 'pypi.use_fallback'")
        use_fallback = asbool(settings['pypi.use_fallback'])
        fallback_mode = 'redirect' if use_fallback else 'none'
    always_show_upstream = settings.get('pypi.always_show_upstream')

    # Using fallback=mirror is the same as fallback=cache and
    # always_show_upstream=true
    if always_show_upstream is None:
        always_show_upstream = fallback_mode == 'mirror'
    else:
        always_show_upstream = asbool(always_show_upstream)
    if fallback_mode == 'mirror':
        LOG.warn("pypi.fallback = mirror is deprecated. You can now use "
                 "pypi.fallback = cache and pypi.always_show_upstream = true")
        fallback_mode = 'cache'

    modes = ('redirect', 'cache', 'none')
    if fallback_mode not in modes:
        raise ValueError("Invalid value for 'pypi.fallback'. "
                         "Must be one of %s" % ', '.join(modes))
    config.registry.fallback = fallback_mode
    config.registry.always_show_upstream = always_show_upstream

    # Special request methods
    config.add_request_method(_app_url, name='app_url')
    config.add_request_method(_locator, name='locator', reify=True)
    config.add_request_method(lambda x: __version__, name='pypicloud_version',
                              reify=True)

    cache_max_age = int(settings.get('pyramid.cache_max_age', 3600))
    config.add_static_view(name='static/%s' % __version__,
                           path='pypicloud:static',
                           cache_max_age=cache_max_age)


def traceback_formatter(excpt, value, tback):
    """ Catches all exceptions and re-formats the traceback raised.
    """

    sys.stdout.write("".join(traceback.format_exception(excpt, value, tback)))


def hook_exceptions():
    """ Hooks into the sys module to set our formatter.
    """

    if hasattr(sys.stdout, "fileno"):  # when testing, sys.stdout is StringIO
        # reopen stdout in non buffered mode
        sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
        # set the hook
        sys.excepthook = traceback_formatter


def main(config, **settings):
    """ This function returns a Pyramid WSGI application.
    """
    hook_exceptions()
    config = Configurator(settings=settings)
    config.include('pypicloud')
    config.scan('pypicloud.views')
    return config.make_wsgi_app()



""" Utilities for authentication and authorization """
import binascii

# pylint: disable=E0611
from paste.httpheaders import AUTHORIZATION, WWW_AUTHENTICATE
# pylint: enable=E0611
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.httpexceptions import HTTPForbidden, HTTPUnauthorized
from pyramid.security import Everyone, authenticated_userid


# Copied from
# http://docs.pylonsproject.org/projects/pyramid_cookbook/en/latest/auth/basic.html
def get_basicauth_credentials(request):
    """ Get the user/password from HTTP basic auth """
    authorization = AUTHORIZATION(request.environ)
    try:
        authmeth, auth = authorization.split(' ', 1)
    except ValueError:  # not enough values to unpack
        return None
    if authmeth.lower() == 'basic':
        try:
            auth = auth.strip().decode('base64')
        except binascii.Error:  # can't decode
            return None
        try:
            login, password = auth.split(':', 1)
        except ValueError:  # not enough values to unpack
            return None
        return {'login': login, 'password': password}

    return None


class BasicAuthenticationPolicy(object):

    """ A :app:`Pyramid` :term:`authentication policy` which
    obtains data from basic authentication headers.

    Constructor Arguments

    ``check``

        A callback passed the credentials and the request,
        expected to return None if the userid doesn't exist or a sequence
        of group identifiers (possibly empty) if the user does exist.
        Required.

    """

    def authenticated_userid(self, request):
        """ Verify login and return the authed userid """
        credentials = get_basicauth_credentials(request)
        if credentials is None:
            return None
        userid = credentials['login']
        if request.access.verify_user(credentials['login'],
                                      credentials['password']):
            return userid
        return None

    def unauthenticated_userid(self, request):
        """ Return userid without performing auth """
        return request.userid

    def effective_principals(self, request):
        """ Get the authed groups for the active user """
        if request.userid is None:
            return [Everyone]
        return request.access.user_principals(request.userid)

    def remember(self, request, principal, **kw):
        """ HTTP Headers to remember credentials """
        return []

    def forget(self, request):
        """ HTTP headers to forget credentials """
        return []


class SessionAuthPolicy(object):

    """ Simple auth policy using beaker sessions """

    def authenticated_userid(self, request):
        """ Return the authenticated userid or ``None`` if no
        authenticated userid can be found. This method of the policy
        should ensure that a record exists in whatever persistent store is
        used related to the user (the user should not have been deleted);
        if a record associated with the current id does not exist in a
        persistent store, it should return ``None``."""
        return request.session.get('user', None)

    def unauthenticated_userid(self, request):
        """ Return the *unauthenticated* userid.  This method performs the
        same duty as ``authenticated_userid`` but is permitted to return the
        userid based only on data present in the request; it needn't (and
        shouldn't) check any persistent store to ensure that the user record
        related to the request userid exists."""
        return request.userid

    def effective_principals(self, request):
        """ Return a sequence representing the effective principals
        including the userid and any groups belonged to by the current
        user, including 'system' groups such as
        ``pyramid.security.Everyone`` and
        ``pyramid.security.Authenticated``. """
        if request.userid is None:
            return [Everyone]
        return request.access.user_principals(request.userid)

    def remember(self, request, principal, **_):
        """
        This implementation is slightly different than expected. The
        application should call remember(userid) rather than
        remember(principal)

        """
        request.session['user'] = principal
        return []

    def forget(self, request):
        """ Return a set of headers suitable for 'forgetting' the
        current user on subsequent requests. """
        request.session.delete()
        return []


def _is_logged_in(request):
    """ Check if there is a logged-in user in the session """
    return request.userid is not None


def _request_login(request):
    """ Return a 401 to force pip to upload its HTTP basic auth credentials """
    response = HTTPUnauthorized()
    realm = WWW_AUTHENTICATE.tuples('Basic realm="%s"' %
                                    request.registry.realm)
    response.headers.update(realm)
    return response


def _forbid(request):
    """
    Return a 403 if user is logged in, otherwise return a 401.

    This is required to force pip to upload its HTTP basic auth credentials

    """
    if request.is_logged_in:
        return HTTPForbidden()
    else:
        return _request_login(request)


def includeme(config):
    """ Configure the app """
    config.set_authorization_policy(ACLAuthorizationPolicy())
    config.set_authentication_policy(config.registry.authentication_policy)
    config.add_authentication_policy(SessionAuthPolicy())
    config.add_authentication_policy(BasicAuthenticationPolicy())
    config.add_request_method(authenticated_userid, name='userid',
                              reify=True)
    config.add_request_method(_forbid, name='forbid')
    config.add_request_method(_request_login, name='request_login')
    config.add_request_method(_is_logged_in, name='is_logged_in', reify=True)

    settings = config.get_settings()
    realm = settings.get('pypi.realm', 'pypi')
    config.registry.realm = realm



""" Store package data in redis """
from datetime import datetime

import json

from .base import ICache


class RedisCache(ICache):

    """ Caching database that uses redis """
    redis_prefix = 'pypicloud:'

    def __init__(self, request=None, db=None, **kwargs):
        super(RedisCache, self).__init__(request, **kwargs)
        self.db = db

    @classmethod
    def configure(cls, settings):
        kwargs = super(RedisCache, cls).configure(settings)
        try:
            from redis import StrictRedis
        except ImportError:  # pragma: no cover
            raise ImportError("You must 'pip install redis' before using "
                              "redis as the database")
        db_url = settings.get('db.url')
        kwargs['db'] = StrictRedis.from_url(db_url)
        return kwargs

    def redis_key(self, key):
        """ Get the key to a redis hash that stores a package """
        return "%spackage:%s" % (self.redis_prefix, key)

    @property
    def redis_set(self):
        """ Get the key to the redis set of package names """
        return self.redis_prefix + 'set'

    def redis_filename_set(self, name):
        """ Get the key to a redis set of filenames for a package """
        return "%sset:%s" % (self.redis_prefix, name)

    def reload_from_storage(self):
        self.clear_all()
        packages = self.storage.list(self.package_class)
        pipe = self.db.pipeline()
        for pkg in packages:
            self.save(pkg, pipe=pipe)
        pipe.execute()

    def fetch(self, filename):
        data = self.db.hgetall(self.redis_key(filename))
        if not data:
            return None
        return self._load(data)

    def _load(self, data):
        """ Load a Package class from redis data """
        name = data.pop('name')
        version = data.pop('version')
        filename = data.pop('filename')
        last_modified = datetime.fromtimestamp(
            float(data.pop('last_modified')))
        kwargs = dict(((k, json.loads(v)) for k, v in data.iteritems()))
        return self.package_class(name, version, filename, last_modified,
                                  **kwargs)

    def all(self, name):
        filenames = self.db.smembers(self.redis_filename_set(name))
        pipe = self.db.pipeline()
        for filename in filenames:
            pipe.hgetall(self.redis_key(filename))
        packages = [self._load(data) for data in pipe.execute()]
        packages.sort(reverse=True)
        return packages

    def distinct(self):
        return list(self.db.smembers(self.redis_set))

    def clear(self, package):
        del self.db[self.redis_key(package.filename)]
        self.db.srem(self.redis_filename_set(package.name), package.filename)
        if self.db.scard(self.redis_filename_set(package.name)) == 0:
            self.db.srem(self.redis_set, package.name)

    def clear_all(self):
        keys = self.db.keys(self.redis_prefix + '*')
        if keys:
            self.db.delete(*keys)

    def save(self, package, pipe=None):
        if pipe is None:
            pipe = self.db
        data = {
            'name': package.name,
            'version': package.version,
            'filename': package.filename,
            'last_modified': package.last_modified.strftime('%s.%f'),
        }
        for key, value in package.data.iteritems():
            data[key] = json.dumps(value)
        pipe.hmset(self.redis_key(package.filename), data)
        pipe.sadd(self.redis_set, package.name)
        pipe.sadd(self.redis_filename_set(package.name), package.filename)



""" Store packages as files on disk """
from datetime import datetime
from contextlib import closing

from pyramid.response import FileResponse

import os
from .base import IStorage
from pypicloud.models import Package


class FileStorage(IStorage):

    """ Stores package files on the filesystem """

    def __init__(self, request=None, **kwargs):
        self.directory = kwargs.pop('directory')
        super(FileStorage, self).__init__(request, **kwargs)

    @classmethod
    def configure(cls, settings):
        kwargs = super(FileStorage, cls).configure(settings)
        directory = os.path.abspath(settings['storage.dir']).rstrip('/')
        if not os.path.exists(directory):
            os.makedirs(directory)
        kwargs['directory'] = directory
        return kwargs

    def get_path(self, package):
        """ Get the fully-qualified file path for a package """
        return os.path.join(self.directory, package.name, package.version,
                            package.filename)

    def list(self, factory=Package):
        for root, _, files in os.walk(self.directory):
            for filename in files:
                shortpath = root[len(self.directory):].strip('/')
                name, version = shortpath.split('/')
                fullpath = os.path.join(root, filename)
                last_modified = datetime.fromtimestamp(os.path.getmtime(
                    fullpath))
                yield factory(name, version, filename, last_modified)

    def download_response(self, package):
        return FileResponse(self.get_path(package),
                            request=self.request,
                            content_type='application/octet-stream')

    def upload(self, package, data):
        destfile = self.get_path(package)
        destdir = os.path.dirname(destfile)
        if not os.path.exists(destdir):
            os.makedirs(destdir)
        uid = os.urandom(4).encode('hex')
        tempfile = os.path.join(destdir, '.' + package.filename + '.' + uid)
        # Write to a temporary file
        with open(tempfile, 'w') as ofile:
            for chunk in iter(lambda: data.read(16 * 1024), ''):
                ofile.write(chunk)

        os.rename(tempfile, destfile)

    def delete(self, package):
        filename = self.get_path(package)
        os.unlink(filename)
        version_dir = os.path.dirname(filename)
        try:
            os.rmdir(version_dir)
        except OSError:
            return
        package_dir = os.path.dirname(version_dir)
        try:
            os.rmdir(package_dir)
        except OSError:
            return

    def open(self, package):
        filename = self.get_path(package)
        return closing(open(filename, 'r'))



""" Views """
from pypicloud.route import Root
from pyramid.view import view_config

from pypicloud import __version__
from pyramid_duh import addslash


@view_config(context=Root, request_method='GET', subpath=(),
             renderer='base.jinja2')
@addslash
def get_index(request):
    """ Render a home screen """
    return {
        'version': __version__,
    }


@view_config(route_name='health', renderer='string')
def health_endpoint(request):
    """ Simple health endpoint """
    return 'OK'



""" Compatibility for different python versions """

# pylint: disable=C0103,C0113
try:
    from functools import total_ordering
except ImportError:  # pragma: no cover
    def total_ordering(cls):
        """Class decorator that fills in missing ordering methods"""
        convert = {
            '__lt__': [
                ('__gt__', lambda self, other: not (self < other or self == other)),
                ('__le__', lambda self, other: self < other or self == other),
                ('__ge__', lambda self, other: not self < other),
            ],
            '__le__': [
                ('__ge__', lambda self, other: not self <= other or self == other),
                ('__lt__', lambda self, other: self <= other and not self == other),
                ('__gt__', lambda self, other: not self <= other),
            ],
            '__gt__': [
                ('__lt__', lambda self, other: not (self > other or self == other)),
                ('__ge__', lambda self, other: self > other or self == other),
                ('__le__', lambda self, other: not self > other),
            ],
            '__ge__': [
                ('__le__', lambda self, other: (not self >= other) or self == other),
                ('__gt__', lambda self, other: self >= other and not self == other),
                ('__lt__', lambda self, other: not self >= other),
            ]
        }
        # Find user-defined comparisons (not those inherited from object).
        roots = [op for op in convert if getattr(
            cls, op, None) is not getattr(object, op, None)]
        if not roots:
            raise ValueError(
                'must define at least one ordering operation: < > <= >=')
        root = max(roots)       # prefer __lt__ to __le__ to __gt__ to __ge__
        for opname, opfunc in convert[root]:
            if opname not in roots:
                opfunc.__name__ = opname
                opfunc.__doc__ = getattr(int, opname).__doc__
                setattr(cls, opname, opfunc)
        return cls



"""LDAP authentication plugin for pypicloud."""


try:
    import ldap
except ImportError:
    raise ImportError(
        "You must 'pip install pypicloud[ldap]' before using ldap as the "
        "authentication backend."
    )

import logging
from functools import wraps

from .base import IAccessBackend


def reconnect(func):
    """
    If the LDAP connection dies underneath us, recreate it
    """
    @wraps(func)
    def _reconnect(*args, **kwargs):
        """
        Inner wrap function to reconnect on failure
        """
        try:
            return func(*args, **kwargs)
        except ldap.LDAPError:
            LDAP._connect()
            return func(*args, **kwargs)

    return _reconnect


class LDAP(object):
    """
    Handles interactions with the remote LDAP server
    """

    @staticmethod
    def configure(settings=None):
        """
        Configures self with the settings dictionary
        """
        LDAP._id_field = settings["auth.ldap.id_field"]
        LDAP._url = settings["auth.ldap.url"]
        LDAP._service_dn = settings["auth.ldap.service_dn"]
        LDAP._service_password = settings["auth.ldap.service_password"]
        LDAP._base_dn = settings["auth.ldap.base_dn"]
        LDAP._all_user_search = settings["auth.ldap.all_user_search"]
        LDAP._admin_field = settings["auth.ldap.admin_field"]
        LDAP._admin_dns = [
            dn for dn in settings["auth.ldap.admin_dns"].splitlines() if dn
        ]
        LDAP._service_account = settings.get("auth.ldap.service_account")

        LDAP._connect()

    @staticmethod
    def _connect():
        """
        Initializes the python-ldap module and does the initial bind
        """
        LDAP._server = ldap.initialize(LDAP._url)
        LDAP._server.simple_bind_s(LDAP._service_dn, LDAP._service_password)

    @staticmethod
    @reconnect
    def _initialize_cache():
        """
        Retrieve the list of all user names and DNs to cache
        """
        results = LDAP._server.search_s(
            LDAP._base_dn,
            ldap.SCOPE_SUBTREE,
            LDAP._all_user_search,
        )
        LDAP._all_users = {}
        if LDAP._service_account:
            LDAP._all_users[LDAP._service_account] = LDAP._service_dn
        for result in results:
            if LDAP._id_field in result[1]:
                LDAP._all_users[result[1][LDAP._id_field][0]] = result[0]

    @staticmethod
    def all_users():
        """
        Returns a list of all user DNs
        """
        if not hasattr(LDAP, "_all_users"):
            LDAP._initialize_cache()
        return list(set(LDAP._all_users.values()))

    @staticmethod
    def all_usernames():
        """
        Returns a list of all user names
        """
        if not hasattr(LDAP, "_all_users"):
            LDAP._initialize_cache()
        return list(set(LDAP._all_users.keys()))

    @staticmethod
    def user_dn(username):
        """
        Returns the dn for the username
        """
        if not hasattr(LDAP, "_all_users"):
            LDAP._initialize_cache()
        return LDAP._all_users.get(username)

    @staticmethod
    @reconnect
    def _add_admins_from_dn(admin_dn):
        """
        Given a DN fragement, add users to _admins and _admin_usernames
        """
        res = LDAP._server.search_s(admin_dn, ldap.SCOPE_SUBTREE)
        try:
            admins = res[0][1][LDAP._admin_field]
        except (IndexError, KeyError) as err:
            logging.warn("Error retrieving admins from %s: %r", admin_dn, err)
            return

        for admin in admins:
            try:
                LDAP._admin_usernames.append(
                    LDAP._server.search_s(
                        admin,
                        ldap.SCOPE_SUBTREE,
                    )[0][1][LDAP._id_field][0]
                )
            except (IndexError, KeyError) as error:
                logging.warn("Error looking up admin %s: %r", admin, error)
            else:
                LDAP._admins.append(admin)

        LDAP._admin_usernames = list(set(LDAP._admin_usernames))
        LDAP._admins = list(set(LDAP._admins))

    @staticmethod
    def admins():
        """
        Returns a list of all the admin DNs
        """
        if not hasattr(LDAP, "_admins"):
            LDAP._admins = [LDAP._service_dn]
            LDAP._admin_usernames = []
            if LDAP._service_account:
                LDAP._admin_usernames.append(LDAP._service_account)
            for admin_dn in LDAP._admin_dns:
                LDAP._add_admins_from_dn(admin_dn)

        return LDAP._admins

    @staticmethod
    def admin_usernames():
        """
        Returns a list of the admin usernames
        """
        if not hasattr(LDAP, "_admin_usernames"):
            LDAP.admins()
        return LDAP._admin_usernames

    @staticmethod
    @reconnect
    def bind_user(user_dn, password):
        """
        Attempts to bind as the user, then rebinds as service user again
        """
        # Empty password may successfully complete an anonymous bind.
        # Explicitly disallow empty passwords.
        if password == "":
            return False

        try:
            LDAP._server.simple_bind_s(user_dn, password)
        except ldap.INVALID_CREDENTIALS:
            return False
        else:
            return True
        finally:
            LDAP._connect()


class LDAPAccessBackend(IAccessBackend):
    """
    This backend allows you to authenticate against a remote LDAP server.
    """

    def __init__(self, request=None, group_map=None, **kwargs):
        super(LDAPAccessBackend, self).__init__(request, **kwargs)
        self.group_map = group_map

    @classmethod
    def configure(cls, settings):
        kwargs = super(LDAPAccessBackend, cls).configure(settings)
        LDAP.configure(settings)
        kwargs["group_map"] = {
            "admin": ("read", "write"),
            "authenticated": ("read",),
        }
        for group in kwargs['default_read']:
            kwargs['group_map'][group] = ('read',)
        for group in kwargs['default_write']:
            kwargs['group_map'][group] = ('read', 'write')
        return kwargs

    def allow_register(self):
        return False

    def _get_password_hash(self, *_):
        return ""

    def verify_user(self, username, password):
        """
        Look up the user DN and attempt to bind with the password
        """
        user_dn = LDAP.user_dn(username)
        return LDAP.bind_user(user_dn, password) if user_dn else False

    def groups(self, username=None):
        """
        Get a list of all groups

        If a username is specified, get all groups that the user belongs to
        """
        if username is None or self.is_admin(username):
            return ["admin"]
        else:
            return []

    def group_members(self, group):
        """
        Get a list of users that belong to a group
        """
        if group == "admin":
            return LDAP.admin_usernames()
        elif group in ("authenticated", "everyone"):
            return LDAP.all_usernames()
        else:
            return []

    def is_admin(self, username):
        """
        Check if the user is an admin
        """
        return username in LDAP.admin_usernames()

    def group_permissions(self, package, group=None):
        """
        Get a mapping of all groups to their permissions on a package

        If a group is specified, return the list of permissions for that group
        """
        if group is None:
            return self.group_map
        else:
            return self.group_map.get(group, [])

    def user_permissions(self, package, username=None):
        """
        Get a mapping of all users to their permissions for a package

        If a username is specified, return a list of permissions for that user
        """
        if username is None:
            return self.group_map
        else:
            perms = set()
            for user_group in self.groups(username):
                perms.update(self.group_permissions(user_group))
            return list(perms)

    def user_package_permissions(self, username):
        """
        Get a list of all packages that a user has permissions to
        """
        all_perms = []
        for package in self.request.db.summary():
            all_perms.append({
                "package": package["name"],
                "permissions": self.user_permissions(package, username),
            })
        return all_perms

    def group_package_permissions(self, group):
        """
        Get a list of all packages that a group has permissions on
        """
        all_perms = []
        for package in self.request.db.summary():
            all_perms.append({
                "package": package["name"],
                "permissions": self.group_permissions(package, group),
            })
        return all_perms

    def user_data(self, username=None):
        """
        Get a list of all users or data for a single user
        """
        if username is None:
            users = []
            for user in LDAP.all_usernames():
                users.append({"username": user, "admin": self.is_admin(user)})
            return users
        else:
            return {
                "username": username,
                "admin": self.is_admin(username),
                "groups": self.groups(username),
            }



""" Views for simple pip interaction """
import posixpath

import logging
import six
from pyramid.httpexceptions import HTTPBadRequest, HTTPFound, HTTPNotFound
from pyramid.view import view_config
from pyramid_duh import argify, addslash

from pypicloud.route import Root, SimplePackageResource, SimpleResource
from pypicloud.util import normalize_name, parse_filename


LOG = logging.getLogger(__name__)


@view_config(context=Root, request_method='POST', subpath=(), renderer='json')
@view_config(context=SimpleResource, request_method='POST', subpath=(),
             renderer='json')
@argify
def upload(request, content, name=None, version=None):
    """ Handle update commands """
    action = request.param(':action', 'file_upload')
    # Direct uploads from the web UI go here, and don't have a name/version
    if name is None or version is None:
        name, version = parse_filename(content.filename)
    else:
        name = normalize_name(name)
    if action == 'file_upload':
        if not request.access.has_permission(name, 'write'):
            return request.forbid()
        try:
            return request.db.upload(content.filename, content.file, name=name,
                                     version=version)
        except ValueError as e:
            return HTTPBadRequest(*e.args)
    else:
        return HTTPBadRequest("Unknown action '%s'" % action)


@view_config(context=SimpleResource, request_method='GET', subpath=(),
             renderer='simple.jinja2')
@addslash
def simple(request):
    """ Render the list of all unique package names """
    names = request.db.distinct()
    i = 0
    while i < len(names):
        name = names[i]
        if not request.access.has_permission(name, 'read'):
            del names[i]
            continue
        i += 1
    return {'pkgs': names}


@view_config(context=SimplePackageResource, request_method='GET', subpath=(),
             renderer='package.jinja2')
@addslash
def package_versions(context, request):
    """ Render the links for all versions of a package """
    fallback = request.registry.fallback
    if fallback == 'redirect':
        if request.registry.always_show_upstream:
            return _simple_redirect_always_show(context, request)
        else:
            return _simple_redirect(context, request)
    elif fallback == 'cache':
        if request.registry.always_show_upstream:
            return _simple_cache_always_show(context, request)
        else:
            return _simple_cache(context, request)
    else:
        return _simple_serve(context, request)


def get_fallback_packages(request, package_name, redirect=True):
    """ Get all package versions for a package from the fallback_url """
    dists = request.locator.get_project(package_name)
    pkgs = {}
    for version, url_set in six.iteritems(dists.get('urls', {})):
        dist = dists[version]
        for url in url_set:
            filename = posixpath.basename(url)
            if not redirect:
                url = request.app_url('api', 'package', dist.name, filename)
            pkgs[filename] = url
    return pkgs


def packages_to_dict(request, packages):
    """ Convert a list of packages to a dict used by the template """
    pkgs = {}
    for package in packages:
        pkgs[package.filename] = package.get_url(request)
    return pkgs


def _pkg_response(pkgs):
    """ Take a package mapping and return either a dict for jinja or a 404 """
    if pkgs:
        return {'pkgs': pkgs}
    else:
        return HTTPNotFound("No packages found")


def _redirect(context, request):
    """ Return a 302 to the fallback url for this package """
    redirect_url = "%s/%s/" % (
        request.registry.fallback_url.rstrip('/'), context.name)
    return HTTPFound(location=redirect_url)


def _simple_redirect(context, request):
    """ Service /simple with fallback=redirect """
    normalized_name = normalize_name(context.name)
    packages = request.db.all(normalized_name)
    if packages:
        if not request.access.has_permission(normalized_name, 'read'):
            if request.is_logged_in:
                return _redirect(context, request)
            else:
                return request.request_login()
        else:
            return _pkg_response(packages_to_dict(request, packages))
    else:
        return _redirect(context, request)


def _simple_redirect_always_show(context, request):
    """ Service /simple with fallback=redirect """
    normalized_name = normalize_name(context.name)
    packages = request.db.all(normalized_name)
    if packages:
        if not request.access.has_permission(normalized_name, 'read'):
            if request.is_logged_in:
                return _redirect(context, request)
            else:
                return request.request_login()
        else:
            pkgs = get_fallback_packages(request, context.name)
            stored_pkgs = packages_to_dict(request, packages)
            # Overwrite existing package urls
            for filename, url in six.iteritems(stored_pkgs):
                pkgs[filename] = url
            return _pkg_response(pkgs)
    else:
        return _redirect(context, request)


def _simple_cache(context, request):
    """ Service /simple with fallback=cache """
    normalized_name = normalize_name(context.name)

    if not request.access.has_permission(normalized_name, 'read'):
        if request.is_logged_in:
            return HTTPNotFound("No packages found named %r" % normalized_name)
        else:
            return request.request_login()

    packages = request.db.all(normalized_name)
    if packages:
        return _pkg_response(packages_to_dict(request, packages))

    if not request.access.can_update_cache():
        if request.is_logged_in:
            return HTTPNotFound("No packages found named %r" % normalized_name)
        else:
            return request.request_login()
    else:
        pkgs = get_fallback_packages(request, context.name, False)
        return _pkg_response(pkgs)


def _simple_cache_always_show(context, request):
    """ Service /simple with fallback=mirror """
    normalized_name = normalize_name(context.name)

    if not request.access.has_permission(normalized_name, 'read'):
        if request.is_logged_in:
            return _redirect(context, request)
        else:
            return request.request_login()

    packages = request.db.all(normalized_name)
    if packages:
        if not request.access.can_update_cache():
            if request.is_logged_in:
                pkgs = get_fallback_packages(request, context.name)
                stored_pkgs = packages_to_dict(request, packages)
                # Overwrite existing package urls
                for filename, url in six.iteritems(stored_pkgs):
                    pkgs[filename] = url
                return _pkg_response(pkgs)
            else:
                return request.request_login()
        else:
            pkgs = get_fallback_packages(request, context.name, False)
            stored_pkgs = packages_to_dict(request, packages)
            # Overwrite existing package urls
            for filename, url in six.iteritems(stored_pkgs):
                pkgs[filename] = url
            return _pkg_response(pkgs)
    else:
        if not request.access.can_update_cache():
            if request.is_logged_in:
                return _redirect(context, request)
            else:
                return request.request_login()
        else:
            pkgs = get_fallback_packages(request, context.name, False)
            return _pkg_response(pkgs)


def _simple_serve(context, request):
    """ Service /simple with fallback=none """
    normalized_name = normalize_name(context.name)

    if not request.access.has_permission(normalized_name, 'read'):
        if request.is_logged_in:
            return HTTPNotFound("No packages found named %r" % normalized_name)
        else:
            return request.request_login()

    packages = request.db.all(normalized_name)
    return _pkg_response(packages_to_dict(request, packages))



""" Setup file """
import sys

from setuptools import setup, find_packages

import os
import re


HERE = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(HERE, 'README.rst')).read()
CHANGES = open(os.path.join(HERE, 'CHANGES.rst')).read()
# Remove custom RST extensions for pypi
CHANGES = re.sub(r'\(\s*:(issue|pr|sha):.*?\)', '', CHANGES)
CHANGES = re.sub(r':ref:`(.*?) <.*>`', r'\1', CHANGES)

REQUIREMENTS = [
    'boto',
    # We're doing enough subclassing and monkey patching to where we really do
    # need to lock this in to a specific version.
    'distlib==0.2.3',
    'paste',
    'passlib',
    'pycrypto',
    'pyramid',
    'pyramid_beaker',
    'pyramid_duh>=0.1.1',
    'pyramid_jinja2',
    'pyramid_tm',
    'rsa',
    'six',
    'transaction',
    'zope.sqlalchemy',
]

TEST_REQUIREMENTS = [
    'flywheel',
    'mock',
    'moto',
    'nose',
    'redis',
    'requests',
    'webtest',
]

if sys.version_info[:2] < (2, 7):
    REQUIREMENTS.append('argparse')
    TEST_REQUIREMENTS.append('unittest2')

if __name__ == "__main__":
    setup(
        name='pypicloud',
        version='0.4.3',
        description='Private PyPI backed by S3',
        long_description=README + '\n\n' + CHANGES,
        classifiers=[
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Development Status :: 4 - Beta',
            'Framework :: Pyramid',
            'Intended Audience :: System Administrators',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Topic :: Internet :: WWW/HTTP',
            'Topic :: System :: Systems Administration',
        ],
        license='MIT',
        author='Steven Arcangeli',
        author_email='stevearc@stevearc.com',
        url='http://pypicloud.readthedocs.org/',
        keywords='pypi s3 cheeseshop package',
        platforms='any',
        zip_safe=False,
        include_package_data=True,
        packages=find_packages(exclude=('tests',)),
        entry_points={
            'console_scripts': [
                'pypicloud-gen-password = pypicloud.scripts:gen_password',
                'pypicloud-make-config = pypicloud.scripts:make_config',
                'ppc-gen-password = pypicloud.scripts:gen_password',
                'ppc-make-config = pypicloud.scripts:make_config',
                'ppc-migrate = pypicloud.scripts:migrate_packages',
                'ppc-export = pypicloud.scripts:export_access',
                'ppc-import = pypicloud.scripts:import_access',
            ],
            'paste.app_factory': [
                'main = pypicloud:main',
            ],
        },
        install_requires=REQUIREMENTS,
        tests_require=REQUIREMENTS + TEST_REQUIREMENTS,
        test_suite='tests',
        extras_require={
            'ldap': ['python-ldap >= 2.4.0'],
            'server': ['waitress'],
            'dynamo': ['flywheel >= 0.2.0'],
        },
    )



""" Store package data in DynamoDB """
from datetime import datetime

import logging
from dynamo3 import DynamoDBConnection
from pkg_resources import parse_version
from pyramid.settings import asbool

from .base import ICache
from pypicloud.models import Package


try:
    from flywheel import Engine, Model, Field, GlobalIndex, __version__
    from flywheel.fields.types import UTC
    if parse_version(__version__) < parse_version('0.2.0'):  # pragma: no cover
        raise ValueError("Pypicloud requires flywheel>=0.2.0")
except ImportError:  # pragma: no cover
    raise ImportError("You must 'pip install flywheel' before using "
                      "DynamoDB as the cache database")

LOG = logging.getLogger(__name__)


class DynamoPackage(Package, Model):

    """ Python package stored in DynamoDB """
    __metadata__ = {
        'global_indexes': [
            GlobalIndex('name-index', 'name'),
        ],
    }
    filename = Field(hash_key=True)
    name = Field()
    version = Field()
    last_modified = Field(data_type=datetime)
    data = Field(data_type=dict)


class PackageSummary(Model):

    """ Aggregate data about packages """
    name = Field(hash_key=True)
    stable = Field()
    unstable = Field()
    last_modified = Field(data_type=datetime)

    def __init__(self, package):
        super(PackageSummary, self).__init__(package.name)
        self.unstable = package.version
        if not package.is_prerelease:
            self.stable = package.version
        self.last_modified = package.last_modified.replace(tzinfo=UTC)

    def update_with(self, package):
        """ Update summary with a package """
        if self.name != package.name:
            LOG.error("Summary name '%s' doesn't match package name '%s'",
                      self.name, package.name)
            return
        self.unstable = max(self.unstable, package.version, key=parse_version)
        self.last_modified = max(self.last_modified.replace(tzinfo=UTC),
                                 package.last_modified.replace(tzinfo=UTC))
        if not package.is_prerelease:
            if self.stable is None:
                self.stable = package.version
            else:
                self.stable = max(self.stable, package.version,
                                  key=parse_version)


class DynamoCache(ICache):

    """ Caching database that uses DynamoDB """
    package_class = DynamoPackage

    def __init__(self, request=None, engine=None, **kwargs):
        super(DynamoCache, self).__init__(request, **kwargs)
        self.engine = engine

    @classmethod
    def configure(cls, settings):
        kwargs = super(DynamoCache, cls).configure(settings)

        access_key = settings.get('db.access_key')
        secret_key = settings.get('db.secret_key')
        region = settings.get('db.region')
        host = settings.get('db.host')
        port = int(settings.get('db.port', 8000))
        secure = asbool(settings.get('db.secure', False))
        namespace = settings.get('db.namespace', ())

        if region is not None:
            connection = DynamoDBConnection.connect(region,
                                                    access_key=access_key,
                                                    secret_key=secret_key)
        elif host is not None:
            connection = DynamoDBConnection.connect('us-east-1',
                                                    host=host,
                                                    port=port,
                                                    is_secure=secure,
                                                    access_key=access_key,
                                                    secret_key=secret_key)
        else:
            raise ValueError("Must specify either db.region or db.host!")
        kwargs['engine'] = engine = Engine(namespace=namespace,
                                           dynamo=connection)

        engine.register(DynamoPackage, PackageSummary)
        engine.create_schema()
        return kwargs

    def fetch(self, filename):
        return self.engine.get(DynamoPackage, filename=filename)

    def all(self, name):
        return sorted(self.engine.query(DynamoPackage).filter(name=name),
                      reverse=True)

    def distinct(self):
        names = set()
        for summary in self.engine.scan(PackageSummary):
            names.add(summary.name)
        return sorted(names)

    def summary(self):
        summaries = sorted(self.engine.scan(PackageSummary),
                           key=lambda s: s.name)
        return [s.__json__() for s in summaries]

    def clear(self, package):
        summary = self.engine.get(PackageSummary, name=package.name)
        if summary is not None and \
            (summary.unstable == package.version or
             summary.stable == package.version or
             summary.last_modified == package.last_modified):
            summary.stable = None
            summary.unstable = '0'
            summary.last_modified = datetime.fromtimestamp(0) \
                .replace(tzinfo=UTC)
            all_packages = self.engine.scan(DynamoPackage)\
                .filter(DynamoPackage.filename != package.filename,
                        name=package.name)
            delete_summary = True
            for package in all_packages:
                delete_summary = False
                summary.update_with(package)
            if delete_summary:
                summary.delete()
            else:
                summary.sync()

        self.engine.delete(package)

    def clear_all(self):
        # We're replacing the schema, so make sure we save and restore the
        # current table/index throughput
        throughput = {}
        for model in (DynamoPackage, PackageSummary):
            tablename = model.meta_.ddb_tablename(self.engine.namespace)
            desc = self.engine.dynamo.describe_table(tablename)
            tablename = model.meta_.ddb_tablename()
            throughput[tablename] = {
                'read': desc.throughput.read,
                'write': desc.throughput.write,
            }
            for index in desc.global_indexes:
                throughput[tablename][index.name] = {
                    'read': index.throughput.read,
                    'write': index.throughput.write,
                }

        self.engine.delete_schema()
        self.engine.create_schema(throughput=throughput)

    def save(self, package):
        summary = self.engine.get(PackageSummary, name=package.name)
        if summary is None:
            summary = PackageSummary(package)
        else:
            summary.update_with(package)

        self.engine.save(package)
        self.engine.sync(summary)



""" Base class for all cache implementations """
from datetime import datetime

import logging
from pkg_resources import parse_version
from pyramid.settings import asbool

import posixpath
from pypicloud.models import Package
from pypicloud.storage import get_storage_impl
from pypicloud.util import parse_filename, normalize_name


LOG = logging.getLogger(__name__)


class ICache(object):

    """ Base class for a caching database that stores package metadata """

    package_class = Package

    def __init__(self, request=None, storage=None, allow_overwrite=None):
        self.request = request
        self.storage = storage(request)
        self.allow_overwrite = allow_overwrite

    def reload_if_needed(self):
        """
        Reload packages from storage backend if cache is empty

        This will be called when the server first starts

        """
        if len(self.distinct()) == 0:
            LOG.info("Cache is empty. Rebuilding from storage backend...")
            self.reload_from_storage()
            LOG.info("Cache repopulated")

    @classmethod
    def configure(cls, settings):
        """ Configure the cache method with app settings """
        return {
            'storage': get_storage_impl(settings),
            'allow_overwrite': asbool(settings.get('pypi.allow_overwrite',
                                                   False)),
        }

    def get_url(self, package):
        """
        Get the download url for a package

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`

        Returns
        -------
        url : str

        """
        return self.storage.get_url(package)

    def download_response(self, package):
        """ Pass through to storage """
        return self.storage.download_response(package)

    def reload_from_storage(self):
        """ Make sure local database is populated with packages """
        self.clear_all()
        packages = self.storage.list(self.package_class)
        for pkg in packages:
            self.save(pkg)

    def upload(self, filename, data, name=None, version=None):
        """
        Save this package to the storage mechanism and to the cache

        Parameters
        ----------
        filename : str
            Name of the package file
        data : file
            File-like readable object
        name : str, optional
            The name of the package (if not provided, will be parsed from
            filename)
        version : str, optional
            The version number of the package (if not provided, will be parsed
            from filename)

        Returns
        -------
        package : :class:`~pypicloud.models.Package`
            The Package object that was uploaded

        Raises
        ------
        e : ValueError
            If the package already exists and allow_overwrite = False

        """
        if version is None:
            name, version = parse_filename(filename, name)
        name = normalize_name(name)
        filename = posixpath.basename(filename)
        old_pkg = self.fetch(filename)
        if old_pkg is not None and not self.allow_overwrite:
            raise ValueError("Package '%s' already exists!" % filename)
        new_pkg = self.package_class(name, version, filename)
        self.storage.upload(new_pkg, data)
        self.save(new_pkg)
        return new_pkg

    def delete(self, package):
        """
        Delete this package from the database and from storage

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`

        """
        self.storage.delete(package)
        self.clear(package)

    def fetch(self, filename):
        """
        Get matching package if it exists

        Parameters
        ----------
        filename : str
            Name of the package file

        Returns
        -------
        package : :class:`~pypicloud.models.Package`

        """
        raise NotImplementedError

    def all(self, name):
        """
        Search for all versions of a package

        Parameters
        ----------
        name : str
            The name of the package

        Returns
        -------
        packages : list
            List of all :class:`~pypicloud.models.Package` s with the given
            name

        """
        raise NotImplementedError

    def distinct(self):
        """
        Get all distinct package names

        Returns
        -------
        names : list
            List of package names

        """
        raise NotImplementedError

    def summary(self):
        """
        Summarize package metadata

        Returns
        -------
        packages : list
            List of package dicts, each of which contains 'name', 'stable',
            'unstable', and 'last_modified'.

        """
        packages = []
        for name in self.distinct():
            pkg = {
                'name': name,
                'stable': None,
                'unstable': '0',
                'last_modified': datetime.fromtimestamp(0),
            }
            for package in self.all(name):
                if not package.is_prerelease:
                    if pkg['stable'] is None:
                        pkg['stable'] = package.version
                    else:
                        pkg['stable'] = max(pkg['stable'], package.version,
                                            key=parse_version)
                pkg['unstable'] = max(pkg['unstable'], package.version,
                                      key=parse_version)
                pkg['last_modified'] = max(pkg['last_modified'],
                                           package.last_modified)
            packages.append(pkg)

        return packages

    def clear(self, package):
        """
        Remove this package from the caching database

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`

        """
        raise NotImplementedError

    def clear_all(self):
        """ Clear all cached packages from the database """
        raise NotImplementedError

    def save(self, package):
        """
        Save this package to the database

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`

        """
        raise NotImplementedError



""" Store packages in S3 """
import calendar
import logging
import posixpath
import time
from contextlib import contextmanager
from hashlib import md5
from urllib import urlopen, quote

import boto.s3
from boto.cloudfront import Distribution
from boto.s3.key import Key
import boto.s3.connection
from pyramid.httpexceptions import HTTPFound
from pyramid.settings import asbool

from .base import IStorage
from pypicloud.models import Package
from pypicloud.util import parse_filename, getdefaults


LOG = logging.getLogger(__name__)

SUPPORTED_CALLING_FORMATS = {
    'SubdomainCallingFormat': boto.s3.connection.SubdomainCallingFormat,
    'VHostCallingFormat': boto.s3.connection.VHostCallingFormat,
    'OrdinaryCallingFormat': boto.s3.connection.OrdinaryCallingFormat,
    'ProtocolIndependentOrdinaryCallingFormat':
        boto.s3.connection.ProtocolIndependentOrdinaryCallingFormat
}


class S3Storage(IStorage):

    """ Storage backend that uses S3 """
    test = False

    def __init__(self, request=None, bucket=None, expire_after=None,
                 bucket_prefix=None, prepend_hash=None, redirect_urls=None,
                 **kwargs):
        super(S3Storage, self).__init__(request, **kwargs)
        self.bucket = bucket
        self.expire_after = expire_after
        self.bucket_prefix = bucket_prefix
        self.prepend_hash = prepend_hash
        self.redirect_urls = redirect_urls

    @classmethod
    def configure(cls, settings):
        kwargs = super(S3Storage, cls).configure(settings)
        kwargs['expire_after'] = int(getdefaults(
            settings, 'storage.expire_after', 'aws.expire_after', 60 * 60 *
            24))
        kwargs['bucket_prefix'] = getdefaults(
            settings, 'storage.prefix', 'aws.prefix', '')
        kwargs['prepend_hash'] = asbool(getdefaults(
            settings, 'storage.prepend_hash', 'aws.prepend_hash', True))
        access_key = getdefaults(settings, 'storage.access_key',
                                 'aws.access_key', None)
        secret_key = getdefaults(settings, 'storage.secret_key',
                                 'aws.secret_key', None)
        host = getdefaults(settings, 'storage.host',
                           'aws.host', boto.s3.connection.NoHostProvided)
        is_secure = getdefaults(settings, 'storage.is_secure',
                                'aws.is_secure', True)
        calling_format = settings.get('storage.calling_format',
                                      'SubdomainCallingFormat')
        kwargs['redirect_urls'] = asbool(settings.get('storage.redirect_urls',
                                                      False))

        if calling_format not in SUPPORTED_CALLING_FORMATS:
            raise ValueError("Only {0} are supported for calling_format"
                             .format(', '.join(SUPPORTED_CALLING_FORMATS)))

        # We used to always use boto.connect_s3 because it can look up buckets
        # in any region. New regions require AWS4-HMAC-SHA256, which boto can
        # only do with a region connection. So if the region is specified (and
        # it must be for new regions like eu-central-1), use a region
        # connection.
        location = settings.get('storage.region')
        if location is None:
            s3conn = boto.connect_s3(
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key,
                host=host,
                is_secure=asbool(is_secure),
                calling_format=SUPPORTED_CALLING_FORMATS[calling_format]())
        else:
            s3conn = boto.s3.connect_to_region(location,
                                               aws_access_key_id=access_key,
                                               aws_secret_access_key=secret_key)
        aws_bucket = getdefaults(settings, 'storage.bucket', 'aws.bucket',
                                 None)
        if aws_bucket is None:
            raise ValueError("You must specify the 'storage.bucket'")
        try:
            bucket = s3conn.get_bucket(aws_bucket)
        except boto.exception.S3ResponseError as e:
            if e.error_code != 'NoSuchBucket':
                if e.status == 301:
                    LOG.warn("Bucket found in different region. Check that "
                             "the S3 bucket specified in 'storage.bucket' is "
                             "in 'storage.region'")
                raise
            location = getdefaults(settings, 'storage.region', 'aws.region',
                                   boto.s3.connection.Location.DEFAULT)
            LOG.info("Creating S3 bucket %s in region %s", aws_bucket,
                     location)
            bucket = s3conn.create_bucket(aws_bucket, location=location)
        kwargs['bucket'] = bucket
        return kwargs

    def calculate_path(self, package):
        """ Calculates the path of a package """
        path = package.name + '/' + package.filename
        if self.prepend_hash:
            m = md5()
            m.update(package.filename)
            prefix = m.digest().encode('hex')[:4]
            path = prefix + '/' + path
        return path

    def get_path(self, package):
        """ Get the fully-qualified bucket path for a package """
        if 'path' not in package.data:
            filename = self.calculate_path(package)
            package.data['path'] = self.bucket_prefix + filename
        return package.data['path']

    def list(self, factory=Package):
        keys = self.bucket.list(self.bucket_prefix)
        for key in keys:
            # Boto doesn't send down metadata from bucket.list()
            # so we are forced to retrieve each key individually.
            key = self.bucket.get_key(key.key)
            filename = posixpath.basename(key.key)
            name = key.get_metadata('name')
            version = key.get_metadata('version')

            # We used to not store metadata. This is for backwards
            # compatibility
            if name is None or version is None:
                try:
                    name, version = parse_filename(filename)
                except ValueError:
                    LOG.warning("S3 file %s has no package name", key.key)
                    continue

            last_modified = boto.utils.parse_ts(key.last_modified)

            pkg = factory(name, version, filename, last_modified, path=key.key)

            yield pkg

    def _generate_url(self, package):
        """ Generate a signed url to the S3 file """
        key = Key(self.bucket, self.get_path(package))
        return key.generate_url(self.expire_after)

    def get_url(self, package):
        if self.redirect_urls:
            return super(S3Storage, self).get_url(package)
        else:
            return self._generate_url(package)

    def download_response(self, package):
        return HTTPFound(location=self._generate_url(package))

    def upload(self, package, data):
        key = Key(self.bucket)
        key.key = self.get_path(package)
        key.set_metadata('name', package.name)
        key.set_metadata('version', package.version)
        # S3 doesn't support uploading from a non-file stream, so we have to
        # read it into memory :(
        key.set_contents_from_string(data.read())

    def delete(self, package):
        path = self.get_path(package)
        key = Key(self.bucket)
        key.key = path
        key.delete()

    @contextmanager
    def open(self, package):
        url = self._generate_url(package)
        handle = urlopen(url)
        try:
            yield handle
        finally:
            handle.close()


class CloudFrontS3Storage(S3Storage):

    """ Storage backend that uses S3 and CloudFront """
    def __init__(self, request=None, bucket=None, expire_after=None, bucket_prefix=None,
                 prepend_hash=None, cloud_front_domain=None, cloud_front_key_file=None,
                 cloud_front_key_string=None, cloud_front_key_id=None, **kwargs):
        super(CloudFrontS3Storage, self).__init__(request, bucket, expire_after, bucket_prefix, prepend_hash, **kwargs)
        self.cloud_front_domain = cloud_front_domain
        self.cloud_front_key_file = cloud_front_key_file
        self.cloud_front_key_id = cloud_front_key_id
        self.cloud_front_key_string = cloud_front_key_string

        self.distribution = Distribution()

    @classmethod
    def configure(cls, settings):
        kwargs = super(CloudFrontS3Storage, cls).configure(settings)
        kwargs['cloud_front_domain'] = getdefaults(
            settings, 'storage.cloud_front_domain', 'aws.cloud_front_domain', '')
        kwargs['cloud_front_key_file'] = getdefaults(
            settings, 'storage.cloud_front_key_file', 'aws.cloud_front_key_file', None)
        kwargs['cloud_front_key_string'] = getdefaults(
            settings, 'storage.cloud_front_key_string', 'aws.cloud_front_key_string', None)
        kwargs['cloud_front_key_id'] = getdefaults(
            settings, 'storage.cloud_front_key_id', 'aws.cloud_front_key_id', '')

        return kwargs

    def _generate_url(self, package):
        """ Get the fully-qualified CloudFront path for a package """
        path = self.calculate_path(package)
        url = self.cloud_front_domain + '/' + quote(path)

        if self.cloud_front_key_file or self.cloud_front_key_string:
            expire_time = int(time.time() + self.expire_after)
            url = self.distribution.create_signed_url(
                url, self.cloud_front_key_id, expire_time, private_key_file=self.cloud_front_key_file,
                private_key_string=self.cloud_front_key_string)

        return url
